"""Status command module."""
import sys

import click
from blessed import Terminal
from colorama import Style

from megalus.main import Megalus
from megalus.status.dashboard import Dashboard


@click.command()
@click.option('--all', is_flag=True)
@click.pass_obj
def status(meg: Megalus, all: bool) -> None:
    """Return docker services status.

    :param all: Show all services
    :param meg: Megalus instance
    :return: None
    """
    dashboard = Dashboard(meg)
    term = Terminal()
    try:
        with term.fullscreen():
            with term.hidden_cursor():
                with term.cbreak():
                    while True:
                        ui = dashboard.get_layout(term, all)
                        ui.display()
                        key_pressed = term.inkey(timeout=5)
                        if 'd' in key_pressed.lower():
                            diff = not diff
                        if 'q' in key_pressed.lower():
                            raise KeyboardInterrupt

    except KeyboardInterrupt:
        print(term.color(0))
        sys.exit(0)
    except BaseException as exc:
        print(term.exit_fullscreen)
        print(Style.RESET_ALL)
        raise exc



"""Check command."""
import os
from typing import List, Optional, Tuple

import arrow
import click
import docker
from docker import DockerClient
from docker.errors import ImageNotFound
from loguru import logger

from megalus.main import Megalus

client = docker.from_env()


def get_services_to_check(meg: Megalus, service: str, service_list: List[str], ignore_list: List[str], tree: List[str],
                          compose_data: dict) -> Optional[Tuple[List, List]]:
    """Get services to check.

    Recursively get all services to check and ignore.

    :param meg: click context object
    :param service: service to inspect
    :param service_list: services to check list
    :param ignore_list: services to ignore list
    :param tree: current service dependencies list
    :param compose_data: docker-compose data
    :return: Tuple
    """
    if service in service_list or service in ignore_list:
        return  # type: ignore

    service_list.append(service) if compose_data.get('build', None) else ignore_list.append(  # type: ignore
        service)
    if compose_data.get('depends_on', None):
        tree = compose_data['depends_on']

    if tree:
        for key in tree:
            key_data = meg.find_service(key)
            get_services_to_check(meg, key, service_list, ignore_list, tree, key_data['compose_data'])
    return service_list, ignore_list


def get_docker_image(compose_data: dict) -> Optional[DockerClient]:  # type: ignore
    """Get docker image data.

    :param compose_data: compose parsed data.
    :return: DockerClient image instance or None
    """
    if compose_data.get('image', None) and compose_data.get('build', None):
        try:
            return client.images.get(compose_data['image'])
        except ImageNotFound:
            return None


def has_old_image(ctx: Megalus, service: str, show_logs) -> bool:
    """Check if image is outdated.

    :param show_logs: Show logs in terminal
    :param ctx: Megalus instance
    :param service: docker service selected
    :return: bool
    """

    def get_date_from_file(file: str) -> arrow:
        """Get date from file.

        :param file: file full path
        :return: Arrow instance
        """
        date = arrow.get(os.path.getmtime(os.path.join(data['working_dir'], file))).to('local')
        if show_logs:
            logger.debug("Last update for file {} was {}".format(file, date.humanize()))
        return date

    data = ctx.find_service(service)
    image = get_docker_image(data['compose_data'])
    if not image:
        return False
    else:
        image_date_created = arrow.get(image.attrs['Created']).to('local')
        if show_logs:
            logger.debug("Last docker build was {}".format(image_date_created.humanize()))
        global_files_to_watch = ctx.config_data['defaults'].get('check_for_build', [])
        list_dates = [
            get_date_from_file(file)
            for file in global_files_to_watch
            if os.path.isfile(os.path.join(data['working_dir'], file))
        ]
        if list_dates and image_date_created < max(list_dates):
            return True
        return False


@click.command()
@click.argument('services', nargs=-1, required=True)
@click.pass_obj
def check(meg: Megalus, services) -> None:
    """Check services.

    This command will check the selected services for:

    * Need build because does not have image
    * Need build because his image is outdated

    To find out if the image needs updating,
    add the list of files in the megalus.yml file,
    whose update date should be compared to the build date of the image.

    Example: if you add the 'Dockerfile', 'requirements.txt' and 'Pipfile'
    inside the 'check_for_build: files:' section in megalus.yml, this command will compare,
    for each file, his last modification date against docker image build date
    for the service.

    :param meg: click context object
    :param services: services list to be inspected
    :return: None
    """

    for service in services:
        logger.info("Checking {}...".format(service))
        service_data = meg.find_service(service)
        logger.debug("Docker Compose file is located at: {}".format(service_data['working_dir']))
        check_docker_image(meg, service_data, show_logs=True)


def check_docker_image(meg, service_data, show_logs=False) -> bool:
    """Check docker image for selected service.

    :param show_logs: Show logs on terminal
    :param meg: Megalus instance
    :param service_data: service data for docker service
    :return: Bool
    """

    need_image = service_data['compose_data'].get('image', None) and service_data['compose_data'].get('build', None)
    if not need_image:
        if show_logs:
            logger.info("Service {} does not need image. Skipping check...".format(service_data['name']))
        return True

    has_image = get_docker_image(service_data['compose_data'])
    if not has_image:
        if show_logs:
            logger.warning("Service {} does not have image.".format(service_data['name']))
        return False

    service_has_old_image = has_old_image(meg, service_data['name'], show_logs=show_logs)
    if service_has_old_image:
        if show_logs:
            logger.warning("Service {} has a old image.".format(service_data['name']))
        return False

    if show_logs:
        logger.success("Service {} pass image check.".format(service_data['name']))
    return True



"""Main module."""
import os
import sys
from typing import Any, Dict, List

import yaml
from buzio import console
from loguru import logger

from megalus.utils import get_path


class ServiceData:
    """Docker Service Data.

    This class represents all data need for the Docker service.

    Attributes
    ==========
        :param name: Service Name (as per `docker-compose.yml` file)
        :param compose: Compose Project Name (as per `megalus.yml` file)
        :param full_name: Unique name for service (docker name and compose project name)
        :param compose_files: Relative path for all `docker-compose.yml` files for this service.
        :param working_dir: Full Working directory for this service
                (which `.env` and main `docker-compose.yml` files resides)
        :param compose_data: Updated compose data for the Service (from all `dock-compose.yaml` files)
        :param compose_project: Compose project data (as per `megalus.yml` file)
        :param all_compose_data: Update Compose Data for all data (from all `dock-compose.yml` files)

    """

    name: str
    compose: str
    full_name: str
    compose_files: List[str]
    working_dir: str
    compose_data: dict
    compose_project: dict
    all_compose_data: dict

    def __init__(self, service_data: dict) -> None:
        """Initialize class.

        :param service_data: dict parsed from Megalus context
        """
        self.name = service_data['name']
        self.compose = service_data['compose']
        self.full_name = service_data['full_name']
        self.compose_files = service_data['compose_files']
        self.working_dir = service_data['compose_data']
        self.compose_data = service_data['compose_data']
        self.compose_project = service_data['compose_project']
        self.all_compose_data = service_data['all_compose_data']


class Megalus:
    """Megalus main class."""

    def __init__(self, config_file: str, logfile: str) -> None:
        """Initialize class.

        :param config_file: path for megalus config path
        :param logfile: path for save log file
        """
        self.service = None
        self._config_file = config_file
        self.base_path = get_path(os.path.dirname(config_file), '.')
        self.compose_data_list = []  # type: List[dict]
        self._data = {}  # type: Dict[str, Any]
        self.all_services = []  # type: List[dict]
        self.all_composes = {}  # type: dict
        self.logfile = logfile
        self._config_data = {}  # type: dict

    @property
    def config_data(self) -> dict:
        """Return megalus configuration data.

        :return: dict
        """
        if self._config_data:
            return self._config_data

        config_path = os.path.join(
            self.base_path,
            os.path.basename(self._config_file)
        )
        try:
            with open(config_path) as file:
                self._config_data = yaml.safe_load(file.read()) or {}
            return self._config_data
        except FileNotFoundError:
            logger.warning("File {} not found. Skipping...".format(config_path))
            return {}

    def _convert_lists(self, data: dict, key: str) -> None:
        """Convert list to dict inside yaml data.

        Works only for Key=Value lists.

        Example:
            environment:
                - DEBUG=false
            ports:
                - "8090:8080"

        Result:
            environment: {"DEBUG": "false"}
            ports: ['8090:8080']

        """
        if isinstance(data[key], list) and "=" in data[key][0]:
            data[key] = {obj.split("=")[0]: obj.split("=")[-1] for obj in data[key]}
        if isinstance(data[key], dict):
            for k in data[key]:
                self._convert_lists(data[key], k)

    def _load_data_from_override(self, source: dict, target: dict, key: str) -> None:
        """Append override data in self.compose.

        Example Compose::
        ---------------
        core:
            build:
                context: ../core
            image: core
            networks:
                - api1
            environment:
                - DEBUG=false
            ports:
             - "8080:80"

        Example override::
        ----------------
        core:
            build:
                dockerfile: Docker_dev
            depends_on:
                - api
            command: bash -c "python manage.py runserver 0.0.0.0"
            environment:
                DEBUG: "True"
            ports:
                - "9000:80"

        Final Result::
        ------------
        core:
            build:
                context: ../core
                dockerfile: Docker_dev
            depends_on:
                - api
            image: core
            command: bash -c "python manage.py runserver 0.0.0.0"
            environment:
                DEBUG: "True"
            networks:
                - api1
            ports:
             - "8080:80"
             - "9000:80"

        """
        if target.get(key, None):
            if isinstance(source[key], dict):
                for k in source[key]:
                    self._load_data_from_override(
                        source=source[key],
                        target=target[key],
                        key=k
                    )
            else:
                if isinstance(target[key], list) and isinstance(source[key], list):
                    target[key] += source[key]
                else:
                    target[key] = source[key]
        else:
            if isinstance(target, list) and isinstance(source[key], list):
                target[key] += source[key]
            else:
                target[key] = source[key]

    def _get_compose_data_for(self, compose_path: str, compose_files: List[str]) -> dict:
        """Read docker compose files data.

        :return: dict
        """
        try:
            resolved_paths = [
                get_path(os.path.join(compose_path, file), base_path=self.base_path)
                for file in compose_files
            ]

            compose_data_list = []
            for compose_file in resolved_paths:
                with open(compose_file, 'r') as file:
                    compose_data = yaml.safe_load(file.read())
                    for key in compose_data:  # type: ignore
                        self._convert_lists(compose_data, key)
                    compose_data_list.append(compose_data)
            reversed_list = list(reversed(compose_data_list))
            self._data = reversed_list[-1]
            for index, override in enumerate(reversed_list):
                self.override = override
                if index + 1 == len(reversed_list):
                    break
                for key in self.override:
                    self._load_data_from_override(self.override, self._data, key)
            return self._data
        except FileNotFoundError:
            logger.warning("File {} not found. Skipping...".format(compose_file or ""))
            return {}

    def get_services(self) -> None:
        """Build service configuration from yaml files.

        :return: None
        """
        if not self.validate_config_data():
            raise SystemExit(1)
        for compose_project in self.config_data.get('compose_projects', []):
            compose_path = self.config_data['compose_projects'][compose_project]['path']
            compose_files = self.config_data['compose_projects'][compose_project]['files']
            compose_data = self._get_compose_data_for(compose_path, compose_files)
            if not compose_data:
                logger.warning("Project {} has errors. Ignoring...".format(compose_project))
                continue
            self.all_composes.update({compose_project: compose_data})
            for service in compose_data['services']:
                self.all_services.append(
                    {
                        'name': service,
                        'compose': compose_project,
                        'full_name': "{} ({})".format(service, compose_project),
                        'compose_files': compose_files,
                        'working_dir': os.path.dirname(
                            get_path(os.path.join(compose_path, compose_files[0]), self.base_path)),
                        'compose_data': compose_data['services'][service],
                        'compose_project': self.config_data['compose_projects'][compose_project],
                        'all_compose_data': compose_data
                    }
                )

    @staticmethod
    def run_command(command: str) -> bool:
        """Run command inside subprocess.

        :param command: string: command to be run
        :return: bool
        """
        logger.debug("Running command: {}".format(command))
        ret = console.run(command)
        if not ret:
            sys.exit(1)
        return ret

    def find_project(self, project_informed: str) -> dict:
        """Find project from string.

        :param project_informed: string informed in command
        :return: compose project data from yml.
        """
        exact_match = list(self.config_data['compose_projects'].keys()).count(project_informed)
        if exact_match == 1:
            return self.config_data['compose_projects'][project_informed]

        eligible_projects = [
            key
            for key in self.config_data['compose_projects'].keys()
            if project_informed in key
        ]
        if not eligible_projects:
            logger.error("Project not found")
            sys.exit(1)

        elif len(eligible_projects) == 1:
            return self.config_data['compose_projects'][eligible_projects[0]]

        project_key = console.choose(eligible_projects, 'Please select project')
        return self.config_data['compose_projects'][project_key]

    def find_service(self, service_informed: str) -> dict:
        """Find service inside megalus service data.

        :param service_informed: string: docker service informed in command.
        :return: docker service megalus data.
        """
        exact_matches = [
            data
            for data in self.all_services
            if data['name'] == service_informed
        ]
        if len(exact_matches) == 1:
            self.service = exact_matches[0]['name']
            return exact_matches[0]

        eligible_services = [
            eligible_service
            for eligible_service in self.all_services
            if service_informed in eligible_service['name']
        ]
        if not eligible_services:
            logger.error("Service not found")
            sys.exit(1)
        elif len(eligible_services) == 1:
            self.service = eligible_services[0]['name']
            return eligible_services[0]
        else:
            choice_list = [
                data['full_name']
                for data in eligible_services
            ]
            service_name = console.choose(choice_list, 'Please select the service')
        data = [
            data
            for data in eligible_services
            if service_name == data['full_name']
        ][0]
        self.service = data['name']
        return data

    def validate_config_data(self) -> bool:
        """Validate data parsed from megalus.yml file."""
        no_errors = True
        if not self.config_data.get('compose_projects'):
            no_errors = False
            logger.error("Key 'compose_projects' not found in configuration file."
                         " Please check and try again.")
        return no_errors



"""Utils Module."""
import datetime
import hashlib
import os
import re
import shutil
from pathlib import Path
from typing import List

import docker
from docker import DockerClient
from loguru import logger

client = docker.from_env()


def find_containers(service: str) -> List[DockerClient]:
    """Find eligible containers for selected service.

    :param service: docker service to run command.
    :return: list of DockerClient containers.
    """
    eligible_containers = [
        container
        for container in client.containers.list()
        if "_{}_".format(service) in container.name
    ]
    return eligible_containers


def backup_folder(folder_path: str) -> None:
    """Backup service context folder.

    :param folder_path: service context folder path
    :return: None
    """
    folders = list(Path(folder_path).parts)
    folder_hash = hashlib.md5(datetime.datetime.now().isoformat().encode("utf-8")).hexdigest()
    last_folder = "{}_{}".format(folders[-1], folder_hash[:10])
    folders.remove(folders[-1])
    folders.append(last_folder)
    backup_path = os.path.join(*folders)
    logger.warning('Directory already exists. Moving folder to {}'.format(backup_path))
    shutil.move(folder_path, backup_path)


def get_path(path: str, base_path: str) -> str:
    """Return real path from string.

    Converts environment variables to path
    Converts relative path to full path
    """

    def _convert_env_to_path(env_in_path):
        s = re.search(r"\${(\w+)}", env_in_path)
        if not s:
            s = re.search(r"(\$\w+)", env_in_path)
        if s:
            env = s.group(1).replace("$", "")
            name = os.environ.get(env)
            if not name:
                raise ValueError("Can't find value for {}".format(env))
            path_list = [
                part if "$" not in part else name
                for part in env_in_path.split("/")
            ]
            path = os.path.join(*path_list)
        else:
            raise ValueError("Cant find path for {}".format(env_in_path))
        return path

    if "$" in base_path:
        base_path = _convert_env_to_path(base_path)
    if "$" in path:
        path = _convert_env_to_path(path)
    if path.startswith("."):
        list_path = os.path.join(base_path, path)
        path = os.path.abspath(list_path)
    return path



"""Command 'run' module."""

import click
from loguru import logger

from megalus.main import Megalus


@click.command()
@click.argument('command', nargs=1, required=True)
@click.pass_obj
def run(meg: Megalus, command: str) -> None:
    """Run selected script.

    :param meg: Megalus instance
    :param command: command/script to execute
    :return: None
    """
    line_to_run = meg.config_data['defaults'].get('scripts', {}).get(command, None)
    if not line_to_run:
        logger.warning('Command "{}" not found in configuration file.'.format(command))
    else:
        meg.run_command(line_to_run)



"""Stop command."""
from typing import List

import click

from megalus.compose.commands import run_compose_command
from megalus.main import Megalus


def stop_all(meg: Megalus) -> None:
    """Stop all containers.

    :param meg: Megalus instance
    :return: None
    """
    all_projects = []  # type: List[dict]
    service_data_list = []
    for service in meg.all_services:
        if service['compose'] not in all_projects:
            service['name'] = ""
            service_data_list.append(service)
            all_projects.append(service['compose'])
    for service_data in service_data_list:
        run_compose_command(meg, "stop", service_data)


@click.command()
@click.argument('services', nargs=-1)
@click.pass_obj
def stop(meg: Megalus, services: List[str]) -> None:
    """Stop selected services.

    :param meg: Megalus instance
    :param services: services to be stopped
    :return: None
    """
    if not services:
        stop_all(meg)
    for service in services:
        service_data = meg.find_service(service)
        run_compose_command(meg, "stop", service_data)



"""Build command."""
from typing import List

import click
from loguru import logger

from megalus.compose.commands import run_compose_command
from megalus.main import Megalus


@click.command()
@click.argument('services', nargs=-1, required=True)
@click.option('--force', is_flag=True)
@click.pass_obj
def build(meg: Megalus, services: List, force: bool) -> None:
    """Run the build command on selected services.

    Use this command to build the selected services.

    :param meg: click context object
    :param services: services list to build
    :param force: use the --force to add the --no-cache option in build
    :return:
    """
    for service_to_find in services:
        logger.info('Looking for Service: {}'.format(service_to_find))
        service_data = meg.find_service(service_to_find)

        options = ["force-rm"]
        if force:
            options.append("no-cache")
        command_args = "| pv -lft -D 2 >> {log}".format(log=meg.logfile)
        run_compose_command(meg, action="build", options=options, service_data=service_data, command_args=command_args)
        logger.success('Service {} builded.'.format(service_data['name']))



"""Megalus Start Command."""
import click

from megalus.compose.commands import run_compose_command
from megalus.main import Megalus
from megalus.utils import get_path


@click.command()
@click.argument('projects', nargs=-1, required=True)
@click.pass_obj
def start(meg: Megalus, projects: str) -> None:
    """Start selected projects.

    :param meg: Megalus instance
    :param projects: Megalus projects to be started up
    :return: None
    """
    for project_informed in projects:
        project_data = meg.find_project(project_informed)
        service_data = {
            'compose_files': project_data['files'],
            'working_dir': get_path(project_data['path'], meg.base_path),
            'compose_project': project_data,
            'all_compose_data': {}
        }
        run_compose_command(
            meg,
            action="up -d",
            service_data=service_data,
            all_services=True
        )



"""Docker-Compose commands module."""
import os
from typing import List, Optional
from urllib.parse import urlparse

import click
import docker
import requests
from loguru import logger

from megalus.main import Megalus

client = docker.from_env()


def get_ngrok_address(service_data: dict, only_domain: bool = False) -> list:
    """Get Ngrok address for selected port.

    :param only_domain: Return only login or full url
    :param service_data: service parsed data
    :return: List
    """
    ngrok_config = service_data['compose_project'].get('ngrok')
    if not ngrok_config:
        return []
    port = ngrok_config['port']
    secure = ngrok_config['secure']
    env = ngrok_config['env']
    protocol = "https" if secure else "http"

    ret = requests.get(
        "http://127.0.0.1:4040/api/tunnels",
        timeout=1
    )
    try:
        ret.raise_for_status()
        api_data = ret.json()
        http_url = [
            obj['public_url']
            for obj in api_data['tunnels']
            if obj['proto'] == protocol and str(port) in obj['config']['addr']
        ]
        if http_url:
            if only_domain:
                return urlparse(http_url[0]).netloc
            else:
                return [
                    "{}={}".format(env, http_url[0])
                ]
        else:
            logger.warning("Ngrok port {} not found. Skipping...".format(port))
            return []
    except requests.exceptions.RequestException as e:
        logger.error("Status Error trying to get ngrok address for port {}: {}".format(port, e))
        return []


def get_env_from_project(service_data: dict) -> list:
    """Get environment variable from yaml.

    Get value from yaml or from memory.

    :param service_data: service parsed data
    :return: List
    """
    project_envs = service_data['compose_project'].get('environment')
    if not project_envs:
        return []
    return [
        "{}={}".format(env, project_envs.get(env, os.getenv(env)))
        for env in project_envs
    ]


def check_for_docker_network(service_data: dict) -> None:
    """Check and create External Docker Networks.

    :param service_data: Dict
    :return: None
    """
    external_network_list = [
        network_name
        for network_name in service_data['all_compose_data'].get('networks', {})
        if service_data['all_compose_data'].get('networks', {}).get(network_name).get('external', False)
    ]
    for external_network in external_network_list:
        network_found = [
            docker_network
            for docker_network in client.networks.list()
            if docker_network.name == external_network
        ]
        if not network_found:
            logger.warning(f"External network '{external_network}' not found. Creating...")
            client.networks.create(name=external_network, driver="bridge")


def run_compose_command(meg: Megalus, action: str, service_data: dict,
                        options: Optional[List[str]] = None, command_args: str = "",
                        all_services: bool = False) -> None:
    """Run docker-compose command.

    :param all_services: The command will be used for all services?
    :param command_args: Command arguments to send after service name in docker-compose command.
    :param meg: Megalus instance
    :param action: docker-compose command
    :param service_data: docker service parsed data
    :param options: docker-compose command options
    :return: None
    """
    environment = []  # type: list
    if service_data['compose_project'].get('ngrok'):
        ngrok_url_env = get_ngrok_address(service_data=service_data)
        ngrok_domain_env = get_ngrok_address(service_data=service_data, only_domain=True)
        environment = ngrok_url_env
        if ngrok_domain_env:
            environment += ["NGROK_DOMAIN={}".format(ngrok_domain_env)]
    environment += get_env_from_project(service_data=service_data)
    check_for_docker_network(service_data)
    meg.run_command(
        "cd {working_dir} && {environment}docker-compose {files} {action}{options}{services}{args}".format(
            working_dir=service_data['working_dir'],
            environment="{} ".format(" ".join(environment)) if environment else "",
            files="-f {}".format(" -f ".join(service_data['compose_files'])),
            options=" --{} ".format(" --".join(options)) if options else " ",
            action=action,
            services=service_data.get('name', "") if not all_services else "",
            args=" {}".format(command_args) if command_args else ""
        )
    )


@click.command()
@click.argument('services', nargs=-1, required=True)
@click.pass_obj
def restart(meg: Megalus, services: List[str]) -> None:
    """Restart selected services.

    :param meg: Megalus instance.
    :param services: Docker services
    :return: None
    """
    for service in services:
        service_data = meg.find_service(service)
        run_compose_command(meg, "restart", service_data)


@click.command()
@click.argument('service', required=True)
@click.argument('number', required=True, default=1, type=click.INT)
@click.pass_obj
def scale(meg: Megalus, service: str, number: int) -> None:
    """Scale selected services.

    :param meg: Megalus instance
    :param service: docker service to be scaled
    :param number: number of replicas
    :return: None
    """
    service_data = meg.find_service(service)
    options = ["scale {}={}".format(service_data['name'], number)]
    run_compose_command(meg, "up -d", options=options, service_data=service_data, all_services=True)


@click.command()
@click.argument('services', nargs=-1, required=True)
@click.option('-d', is_flag=True)
@click.pass_obj
def up(meg: Megalus, services: List[str], d: bool) -> None:
    """Start selected services.

    :param meg: Megalus instance
    :param services: Services to be started up
    :param d: detached option
    :return: None
    """
    command = "up -d" if d or len(services) > 1 else "up"
    for service in services:
        service_data = meg.find_service(service)
        run_compose_command(meg, command, service_data=service_data)



"""Docker-Compose down module."""
from typing import List

import click

from megalus.compose.commands import run_compose_command
from megalus.main import Megalus
from megalus.utils import get_path


@click.command()
@click.option('--remove-all', is_flag=True)
@click.argument('projects', nargs=-1)
@click.pass_obj
def down(meg: Megalus, projects: List[str], remove_all: bool) -> None:
    """Down compose project.

    :param meg: Megalus instance
    :param projects: projects to down
    :param remove_all: remove image and containers too
    :return: None
    """
    options = ["rmi all", "volumes", "remove-orphans"] if remove_all else None

    projects_to_run = projects if projects else meg.config_data['compose_projects'].keys()

    for project_informed in projects_to_run:
        project_data = meg.find_project(project_informed)
        service_data = {
            'compose_files': project_data['files'],
            'working_dir': get_path(project_data['path'], meg.base_path),
            'compose_project': project_data,
            'all_compose_data': {}
        }
        run_compose_command(
            meg,
            action="down",
            service_data=service_data,
            options=options,
            all_services=True
        )



"""Test bash command."""
from click.testing import CliRunner

from megalus.bash.commands import bash


def test_bash_with_service(caplog, obj, mocker):
    runner = CliRunner()
    with runner.isolated_filesystem():
        mocker.patch('megalus.main.console.run')
        mocker.patch('megalus.bash.commands.find_containers', return_value=[])
        result = runner.invoke(bash, ['django'], obj=obj)
        running_command = [
            message
            for message in caplog.messages
            if "Running command:" in message
        ][0]
        service_path = [
            service['working_dir']
            for service in obj.all_services
            if service['name'] == 'django'
        ][0]
        assert result.exit_code == 0
        assert 'cd {} && docker-compose -f docker-compose.yml -f docker-compose.override.yml ' \
               'run --rm --service-ports django /bin/bash'.format(service_path) in running_command


def test_bash_with_container(caplog, obj, mocker, django_container):
    runner = CliRunner()
    with runner.isolated_filesystem():
        mocker.patch('megalus.main.console.run')
        mocker.patch('megalus.bash.commands.find_containers', return_value=[django_container])
        result = runner.invoke(bash, ['django'], obj=obj)
        running_command = [
            message
            for message in caplog.messages
            if "Running command:" in message
        ][0]
        service_path = [
            service['working_dir']
            for service in obj.all_services
            if service['name'] == 'django'
        ][0]
        assert result.exit_code == 0
        assert 'cd {} && docker exec -ti {} /bin/bash'.format(
            service_path, django_container.short_id) in running_command



"""Test run command."""
from click.testing import CliRunner

from megalus.run.commands import run


def test_update_script(caplog, obj, mocker):
    runner = CliRunner()
    with runner.isolated_filesystem():
        mocker.patch('megalus.main.console.run')
        result = runner.invoke(run, ['update'], obj=obj)
        assert result.exit_code == 0
        running_command = [
            message
            for message in caplog.messages
            if "Running command:" in message
        ][0]
        assert 'make update' in running_command
